@@ -828,6 +828,14 @@ class test_driver(unittest.TestCase):
 	def testCursorRead(self):
 		self.cursor_read()
 
+	@pg_tmp
+	def testCursorIter(self):
+		ps = db.prepare("SELECT i FROM generate_series(0, 10) AS g(i)")
+		c = ps.declare()
+		self.failUnlessEqual(next(iter(c)), (0,))
+		self.failUnlessEqual(next(iter(c)), (1,))
+		self.failUnlessEqual(next(iter(c)), (2,))
+
 	@pg_tmp
 	def testCursorReadInXact(self):
 		with db.xact():
@@ -879,6 +887,22 @@ class test_driver(unittest.TestCase):
 		c.seek(10, 2)
 		self.failUnlessEqual(r1, c.read(10))
 
+	@pg_tmp
+	def testSeek(self):
+		ps = db.prepare("SELECT i FROM generate_series(0, (2^6)::int - 1) AS g(i)")
+		c = ps.declare()
+
+		self.failUnlessEqual(c.seek(4, 'FORWARD'), 4)
+		self.failUnlessEqual([x for x, in c.read(10)], list(range(4, 14)))
+
+		self.failUnlessEqual(c.seek(2, 'BACKWARD'), 2)
+		self.failUnlessEqual([x for x, in c.read(10)], list(range(12, 22)))
+
+		self.failUnlessEqual(c.seek(-5, 'BACKWARD'), 5)
+		self.failUnlessEqual([x for x, in c.read(10)], list(range(27, 37)))
+
+		self.failUnlessEqual(c.seek('ALL'), 27)
+
 	def testScrollBackwards(self):
 		# testScroll again, but backwards this time.
 		self.testScroll(direction = False)
@@ -1078,7 +1102,35 @@ class test_driver(unittest.TestCase):
 				)
 
 	@pg_tmp
-	def testXML(self):
+	def testDomainSupport(self):
+		'test domain type I/O'
+
+		db.execute('CREATE DOMAIN int_t AS int')
+		db.execute('CREATE DOMAIN int_t_2 AS int_t')
+		db.execute('CREATE TYPE tt AS (a int_t, b int_t_2)')
+
+		samples = {
+			'int_t': [10],
+			'int_t_2': [11],
+			'tt': [(12, 13)]
+		}
+
+		for (typname, sample_data) in samples.items():
+			pb = db.prepare(
+				"SELECT $1::" + typname
+			)
+			for sample in sample_data:
+				rsample = list(pb.rows(sample))[0][0]
+				if isinstance(rsample, pg_types.Array):
+					rsample = rsample.nest()
+				self.failUnless(
+					rsample == sample,
+					"failed to return %s object data as-is; gave %r, received %r" %(
+						typname, sample, rsample
+					)
+				)
+
+	def check_xml(self):
 		try:
 			xml = db.prepare('select $1::xml')
 			textxml = db.prepare('select $1::text::xml')
@@ -1088,7 +1140,13 @@ class test_driver(unittest.TestCase):
 			return
 		foo = etree.XML('<foo/>')
 		bar = etree.XML('<bar/>')
-		tostr = etree.tostring
+		if hasattr(etree, 'tostringlist'):
+			# 3.2
+			def tostr(x):
+				return etree.tostring(x, encoding='utf-8')
+		else:
+			# 3.1 compat
+			tostr = etree.tostring
 		self.failUnlessEqual(tostr(xml.first(foo)), tostr(foo))
 		self.failUnlessEqual(tostr(xml.first(bar)), tostr(bar))
 		self.failUnlessEqual(tostr(textxml.first('<foo/>')), tostr(foo))
@@ -1125,6 +1183,23 @@ class test_driver(unittest.TestCase):
 			(tostr(foo), tostr(bar))
 		)
 
+	@pg_tmp
+	def testXML(self):
+		self.check_xml()
+
+	@pg_tmp
+	def testXML_ascii(self):
+		# check a non-utf8 encoding (3.2 and up)
+		db.settings['client_encoding'] = 'sql_ascii'
+		self.check_xml()
+
+	@pg_tmp
+	def testXML_utf8(self):
+		# in 3.2 we always serialize at utf-8, so check that
+		# that path is being ran by forcing the client_encoding to utf8.
+		db.settings['client_encoding'] = 'utf8'
+		self.check_xml()
+
 	@pg_tmp
 	def testUUID(self):
 		# doesn't exist in all versions supported by py-postgresql.
@@ -1510,6 +1585,23 @@ class test_driver(unittest.TestCase):
 		db.execute("drop table subxact_sx2")
 		db.execute("drop table subxact_sx3")
 
+	@pg_tmp
+	def testReleasedSavepoint(self):
+		# validate that the rolled back savepoint is released as well.
+		x = None
+		with db.xact():
+			try:
+				with db.xact():
+					try:
+						with db.xact() as x:
+							db.execute("selekt 1")
+					except pg_exc.SyntaxError:
+						db.execute('RELEASE "xact(' + hex(id(x)) + ')"')
+			except pg_exc.InvalidSavepointSpecificationError as e:
+				pass
+			else:
+				self.fail("InvalidSavepointSpecificationError not raised")
+
 	@pg_tmp
 	def testCloseInSubTransactionBlock(self):
 		try:
